#include <math.h>
#include <stdio.h>
#include <sys/stat.h>

#include "http_loop.h"

void http_loop_res_set_body_buf(http_loop_ctx *ctx, const char *body,
                                size_t body_len, bool owned) {
  ctx->res.body_type = http_response_body_buf;
  ctx->res.body = (void *)body;
  ctx->res.body_len = body_len;
  ctx->res.owns_body = owned;
}

void http_loop_res_set_body_file(http_loop_ctx *ctx, const char *filename) {
  struct stat st;
  stat(filename, &st);

  // TODO error handling
  FILE *f = fopen(filename, "r");

  ctx->res.body_type = http_response_body_file;
  ctx->res.body = f;
  ctx->res.body_len = st.st_size;
}

void http_loop_res_add_header(http_loop_ctx *ctx, http_header type,
                              const char *value, bool owned) {
  ctx->res.headers[ctx->res.header_count].type = type;
  ctx->res.headers[ctx->res.header_count].value = value;
  ctx->res.headers[ctx->res.header_count].owned = owned;

  ctx->res.header_count++;
}

/*
 * Converts a string to a number, returning true if the string contained a valid
 * positive number.
 */
static bool string_to_num(size_t *res, const char *s, size_t len) {
  *res = 0;

  for (size_t i = 0; i < len; i++) {
    int val = s[i] - '0';

    if (val < 0 || val > 9) {
      return false;
    }

    *res += val * (int)pow(10, (len - 1) - i);
  }

  return true;
}

bool http_loop_step_body_to_buf(event_loop_conn *conn) {
  http_loop_ctx *ctx = conn->ctx;

  if (ctx->req.body_len == 0) {
    for (size_t i = 0; i < ctx->req.num_headers; i++) {
      struct phr_header *header = &ctx->req.headers[i];

      if (strncmp(header->name, "Content-Length", header->name_len) == 0) {
        // If the content length header is present but contains an invalid
        // number, we return a bad request error
        if (!string_to_num(&ctx->req.body_len, header->value,
                           header->value_len)) {
          ctx->res.status = http_bad_request;
          conn->state = event_loop_conn_state_res;

          return true;
        }
        // The content length was actually 0, so we can instantly return here
        else if (ctx->req.body_len == 0) {
          return true;
        }
      }
    }

    // A zero here means there's no content length header
    if (ctx->req.body_len == 0) {
      ctx->res.status = http_length_required;
      conn->state = event_loop_conn_state_res;

      return true;
    }

    ctx->req.body = malloc(ctx->req.body_len * sizeof(uint8_t));
    ctx->req.body_received = 0;
  }

  size_t bytes_to_copy = MIN(conn->rbuf_size - conn->rbuf_read,
                             ctx->req.body_len - ctx->req.body_received);
  memcpy(&ctx->req.body[ctx->req.body_received], &conn->rbuf[conn->rbuf_read],
         bytes_to_copy);
  ctx->req.body_received += bytes_to_copy;

  return ctx->req.body_received == ctx->req.body_len;
}

bool http_loop_step_auth(event_loop_conn *conn) {
  http_loop_ctx *ctx = conn->ctx;

  for (size_t i = 0; i < ctx->req.num_headers; i++) {
    struct phr_header *header = &ctx->req.headers[i];

    if ((strncmp("X-Api-Key", header->name, header->name_len) == 0) &&
        (strncmp(header->value, ctx->g->api_key, header->value_len) == 0)) {
      return true;
    }
  }

  ctx->res.status = http_unauthorized;
  conn->state = event_loop_conn_state_res;

  return true;
}
